"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const command_1 = __importStar(require("@oclif/command"));
const npm_auth_1 = require("../../lib/npm-auth");
const Semver = __importStar(require("../../lib/semver"));
const context_1 = require("../../utils/context");
const context_checkers_1 = require("../../utils/context-checkers");
const contrext_guard_1 = require("../../utils/contrext-guard");
const Output = __importStar(require("../../utils/output"));
const Publish = __importStar(require("../../utils/publish"));
const Rel = __importStar(require("../../utils/release"));
class Preview extends command_1.default {
    run() {
        return __awaiter(this, void 0, void 0, function* () {
            const { flags } = this.parse(Preview);
            // TODO validate for missing or faulty package.json
            // TODO validate for dirty git status
            // TODO validate for found releases that fall outside the subset we support.
            // For example #.#.#-foobar.1 is something we would not know what to do
            // with. A good default is probably to hard-error when these are
            // encountered. But offer a flag/config option called e.g. "ignore_unsupport_pre_release_identifiers"
            // TODO handle edge case: not a git repo
            // TODO handle edge case: a git repo with no commits
            // TODO nicer tag rendering:
            //    1. for annotated tags should the messge
            //    2. show the tag author name
            //    3. show the the date the tag was made
            const context = yield context_1.getContext({
                overrides: {
                    trunk: flags.trunk || null,
                },
            });
            const report = contrext_guard_1.check({ context })
                .errorUnless(context_checkers_1.npmAuthSetup())
                .errorUnless(context_checkers_1.isTrunk())
                .errorUnless(notAlreadyStableOrPreviewReleased())
                .stopUnless(haveCommitsInTheSeries())
                .stopUnless(haveMeaningfulCommitsInTheSeries())
                .run();
            const maybeRelease = Rel.getNextPreview(context.series);
            if (flags['build-num'] && !Rel.isNoReleaseReason(maybeRelease)) {
                maybeRelease.version = Semver.setBuildNum(maybeRelease.version, flags['build-num']);
            }
            if (flags['dry-run']) {
                return Output.outputOk('dry_run', {
                    report: report,
                    release: maybeRelease,
                    commits: context.series.commitsInNextPreview.map(c => c.message),
                });
            }
            if (report.errors.length) {
                contrext_guard_1.guard({ context, report, json: flags.json });
            }
            if (flags.json && report.stops.length) {
                Output.didNotPublish({ reasons: report.stops });
                return this.exit(0);
            }
            const release = maybeRelease; // now validated
            const publishPlan = {
                release: {
                    distTag: 'next',
                    version: release.version.version,
                },
                options: {
                    skipNPM: flags['skip-npm'],
                    showProgress: !flags.json,
                },
            };
            npm_auth_1.setupNPMAuthfileOnCI();
            yield Publish.publish(publishPlan);
            if (flags.json) {
                Output.didPublish({ release: publishPlan.release });
            }
        });
    }
}
exports.Preview = Preview;
Preview.flags = {
    trunk: command_1.flags.string({
        default: '',
        description: 'State which branch is trunk. Defaults to honuring the "base" branch setting in the GitHub repo settings.',
    }),
    ['build-num']: command_1.flags.integer({
        description: 'Force a build number. Should not be needed generally. For exceptional cases.',
        char: 'n',
    }),
    'dry-run': command_1.flags.boolean({
        default: false,
        description: 'output what the next version would be if released now',
        char: 'd',
    }),
    'skip-npm': command_1.flags.boolean({
        default: false,
        description: 'skip the step of publishing the package to npm',
    }),
    json: command_1.flags.boolean({
        default: false,
        description: 'format output as JSON',
        char: 'j',
    }),
};
//
// Validators
//
function haveCommitsInTheSeries() {
    return {
        code: 'series_empty',
        summary: 'A preview release must have at least one commit since the last preview',
        run(ctx) {
            const release = Rel.getNextPreview(ctx.series);
            return release !== 'empty_series';
        },
    };
}
function haveMeaningfulCommitsInTheSeries() {
    return {
        code: 'series_only_has_meaningless_commits',
        summary: 'A preview release must have at least one semantic commit',
        run(ctx) {
            const release = Rel.getNextPreview(ctx.series);
            return release !== 'no_meaningful_change';
            // todo
            // hint:   //             'All commits are either meta or not conforming to conventional commit. No release will be made.',
        },
    };
}
function notAlreadyStableOrPreviewReleased() {
    return {
        code: 'preview_on_commit_with_preview_and_or_stable',
        summary: 'A preview release requires the commit to have no existing stable or preview release.',
        run(ctx) {
            if (ctx.series.current.releases.stable &&
                ctx.series.current.releases.preview) {
                return { kind: 'fail', details: { subCode: 'preview_and_stable' } };
            }
            if (ctx.series.current.releases.stable) {
                return { kind: 'fail', details: { subCode: 'stable' } };
            }
            if (ctx.series.current.releases.preview) {
                return { kind: 'fail', details: { subCode: 'preview' } };
            }
            return true;
        },
    };
}
