"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * This module deals with building changelogs from a serieses.
 */
const chalk_1 = __importDefault(require("chalk"));
const common_tags_1 = require("common-tags");
const utils_1 = require("./utils");
function empty() {
    return {
        breaking: { commits: [], label: 'BREAKING CHANGES' },
        features: { commits: [], label: 'Features' },
        chores: { commits: [], label: 'Chores' },
        fixes: { commits: [], label: 'Fixes' },
        improvements: { commits: [], label: 'Improvements' },
        unspecified: { commits: [], label: 'Unspecified Changes' },
    };
}
exports.empty = empty;
function organize(series) {
    const log = empty();
    for (const c of series.commitsInNextStable) {
        if (c.message.parsed === null) {
            log.unspecified.commits.push(c);
            continue;
        }
        const cp = c.message.parsed;
        // breaking changes are collected as a group in addition to by type.
        if (cp.breakingChange) {
            log.breaking.commits.push(c);
        }
        if (cp.typeKind === 'feat') {
            log.features.commits.push(c);
        }
        else if (cp.typeKind === 'fix') {
            log.fixes.commits.push(c);
        }
        else if (cp.typeKind === 'chore') {
            log.chores.commits.push(c);
        }
        else if (cp.typeKind === 'other') {
            log.improvements.commits.push(c);
        }
        else {
            utils_1.casesHandled(cp.typeKind);
        }
    }
    return log;
}
function render(series, opts) {
    if (opts.type === 'markdown')
        return Markdown.render(organize(series));
    if (opts.type === 'plain')
        return Plain.render(organize(series));
    utils_1.casesHandled(opts.type);
}
exports.render = render;
var Plain;
(function (Plain) {
    function render(log) {
        const order = [
            'breaking',
            'features',
            'fixes',
            'improvements',
            'chores',
        ];
        const doc = order
            .filter(sectionName => {
            return log[sectionName].commits.length > 0;
        })
            .map(sectionName => {
            if (sectionName === 'breaking') {
                return (sectionTitle(log[sectionName].label) +
                    '\n\n' +
                    sectionCommits(log[sectionName].commits, { breaking: false }) +
                    '\n');
            }
            if (sectionName === 'improvements') {
                return (sectionTitle(log[sectionName].label) +
                    '\n\n' +
                    sectionCommits(log[sectionName].commits, { type: true }) +
                    '\n');
            }
            return (sectionTitle(log[sectionName].label) +
                '\n\n' +
                sectionCommits(log[sectionName].commits) +
                '\n');
        });
        if (log.unspecified.commits.length) {
            doc.push(sectionTitle(log.unspecified.label) +
                '\n\n' +
                '  ' +
                log.unspecified.commits
                    .map(c => `${chalk_1.default.gray(shortSha(c))} ${c.message.raw}`)
                    .join('\n  ') +
                '\n');
        }
        return doc.join('\n');
    }
    Plain.render = render;
    function sectionCommits(cs, opts) {
        return cs.map(c => sectionCommit(c, opts)).join('\n');
    }
    function sectionTitle(title) {
        return chalk_1.default.magenta(title);
    }
    function sectionCommit(c, opts) {
        var _a, _b;
        const sha = chalk_1.default.gray(shortSha(c));
        const type = ((_a = opts) === null || _a === void 0 ? void 0 : _a.type) === true ? ' ' + c.message.parsed.type + ':' : '';
        const description = ' ' + c.message.parsed.description;
        const breaking = ((_b = opts) === null || _b === void 0 ? void 0 : _b.breaking) === false
            ? ''
            : c.message.parsed.breakingChange
                ? chalk_1.default.red(' (breaking)')
                : '';
        return `  ${sha}${breaking}${type}${description}`;
    }
})(Plain || (Plain = {}));
var Markdown;
(function (Markdown) {
    function render(log) {
        const order = [
            'breaking',
            'features',
            'fixes',
            'improvements',
            'chores',
        ];
        const doc = order
            .filter(sectionName => {
            return log[sectionName].commits.length > 0;
        })
            .map(sectionName => {
            if (sectionName === 'breaking') {
                return (common_tags_1.stripIndents `
            ${sectionTitle(log[sectionName].label)}
  
            ${log[sectionName].commits
                    .map(c => sectionCommit(c, { breaking: false }))
                    .join('\n')}
          ` + '\n');
            }
            if (sectionName === 'improvements') {
                return (common_tags_1.stripIndents `
            ${sectionTitle(log[sectionName].label)}
  
            ${log[sectionName].commits
                    .map(c => sectionCommit(c, { type: true }))
                    .join('\n')}
          ` + '\n');
            }
            return (common_tags_1.stripIndents `
          ${sectionTitle(log[sectionName].label)}

          ${sectionCommits(log[sectionName].commits)}
        ` + '\n');
        });
        if (log.unspecified.commits.length) {
            doc.push(common_tags_1.stripIndents `
        ${sectionTitle(log.unspecified.label)}

        - ${log.unspecified.commits
                .map(c => `${shortSha(c)} ${c.message.raw}`)
                .join('\n- ')}
      ` + '\n');
        }
        return doc.join('\n');
    }
    Markdown.render = render;
    function sectionCommits(cs) {
        return cs.map(c => sectionCommit(c)).join('\n');
    }
    function sectionTitle(title) {
        return `#### ${title}`;
    }
    function sectionCommit(c, opts) {
        var _a, _b;
        const sha = shortSha(c);
        const type = ((_a = opts) === null || _a === void 0 ? void 0 : _a.type) === true ? ' ' + c.message.parsed.type + ':' : '';
        const description = ' ' + c.message.parsed.description;
        const breaking = ((_b = opts) === null || _b === void 0 ? void 0 : _b.breaking) === false
            ? ''
            : c.message.parsed.breakingChange
                ? ' (breaking)'
                : '';
        return `- ${sha}${breaking}${type}${description}`;
    }
})(Markdown || (Markdown = {}));
function shortSha(c) {
    return c.sha.slice(0, 7);
}
