"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Given a list of conventional commit messages (subject and body, the entire
 * message for the commit) calculate what the package version containing these
 * changes should be. Returns `null` if all changes were meta or unconforming.
 */
function calcBumpType(isInitialDevelopment, commitMessages) {
    let semverPart = null;
    for (const m of commitMessages) {
        const cc = parse(m);
        // Commits that do not conform to conventional commit standard are discarded
        if (!cc)
            continue;
        // Completing initial development is a spec extension
        // https://github.com/conventional-commits/conventionalcommits.org/pull/214
        if (isInitialDevelopment && cc.completesInitialDevelopment)
            return 'major';
        if (isMetaChange(cc)) {
            // chore type commits are considered to not change the runtime in any way
            continue;
        }
        // Nothing can be be higher so we've reached our final value effectively.
        if (cc.breakingChange) {
            // during initial development breaking changes are permitted without
            // having to bump the major.
            semverPart = isInitialDevelopment ? 'minor' : 'major';
            break;
        }
        // If already at minor continue, now looking only for major changes
        if (semverPart === 'minor') {
            continue;
        }
        if (isMinorChange(cc)) {
            semverPart = 'minor';
            // during initial development breaking changes are permitted without
            // having to bump the major. Therefore, we know we won't get a bunmpType
            // higher than this, can short-circuit.
            if (isInitialDevelopment)
                break;
            else
                continue;
        }
        semverPart = 'patch';
    }
    return semverPart;
}
exports.calcBumpType = calcBumpType;
function isMinorChange(conventionalCommit) {
    return ['feat', 'feature'].includes(conventionalCommit.type);
}
function isMetaChange(conventionalCommit) {
    return ['chore'].includes(conventionalCommit.type);
}
const pattern = /^([^:\r\n(]+)(?:\(([^\r\n()]+)\))?:\s*([^\r\n]+)[\n\r]*(.*)$/s;
function parse(message) {
    var _a, _b, _c, _d, _e, _f;
    const result = message.match(pattern);
    if (!result)
        return null;
    const [, type, scope, description, rest] = result;
    let completesInitialDevelopment = false;
    let breakingChange = null;
    let body = null;
    let footers = [];
    if (rest) {
        const rawFooters = [];
        let currFooter = -1;
        let currSection = 'body';
        for (const para of rest.split(/(?:\r?\n){2}/)) {
            if (para.match(/^COMPLETES[-\s]INITIAL[-\s]DEVELOPMENT\s*/)) {
                completesInitialDevelopment = true;
            }
            else if (para.match(/^\s*BREAKING[-\s]CHANGE\s*:\s*.*/)) {
                currSection = 'breaking_change';
                breakingChange =
                    ((breakingChange !== null && breakingChange !== void 0 ? breakingChange : '')) +
                        '\n\n' +
                        para.replace(/^BREAKING[-\s]CHANGE\s*:/, '');
            }
            else if (para.match(/^\s*[\w-]+\s*:.*/)) {
                currSection = 'footers';
                rawFooters.push(para);
                currFooter++;
            }
            else if (currSection === 'body') {
                body = ((body !== null && body !== void 0 ? body : '')) + '\n\n' + para;
            }
            else if (currSection === 'breaking_change') {
                breakingChange = ((breakingChange !== null && breakingChange !== void 0 ? breakingChange : '')) + '\n\n' + para;
            }
            else {
                rawFooters[currFooter] += '\n\n' + para;
            }
        }
        footers = rawFooters.map(f => {
            const [, type, body] = f.trim().split(/^\s*([\w-]+)\s*:/);
            return {
                type: type.trim(),
                body: body.trim(),
            };
        });
    }
    const typeTrimmed = type.trim();
    return {
        typeKind: getKind(typeTrimmed),
        type: typeTrimmed,
        scope: (_b = (_a = scope) === null || _a === void 0 ? void 0 : _a.trim(), (_b !== null && _b !== void 0 ? _b : null)),
        description: description.trim(),
        body: (_d = (_c = body) === null || _c === void 0 ? void 0 : _c.trim(), (_d !== null && _d !== void 0 ? _d : null)),
        footers: (footers !== null && footers !== void 0 ? footers : []),
        breakingChange: (_f = (_e = breakingChange) === null || _e === void 0 ? void 0 : _e.trim(), (_f !== null && _f !== void 0 ? _f : null)),
        completesInitialDevelopment,
    };
}
exports.parse = parse;
function getKind(s) {
    if (s === 'feat')
        return 'feat';
    if (s === 'fix')
        return 'fix';
    if (s === 'chore')
        return 'chore';
    return 'other';
}
