"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const child_process_1 = require("child_process");
const common_tags_1 = require("common-tags");
const Path = __importStar(require("path"));
/**
 * A wrapper around spawn, easier to use.
 */
function run(commandRaw, options) {
    var _a;
    return __awaiter(this, void 0, void 0, function* () {
        const command = parseCommandString(commandRaw);
        const env = ((_a = options) === null || _a === void 0 ? void 0 : _a.envAdditions) ? Object.assign(Object.assign({}, process.env), options.envAdditions) : process.env;
        const child = child_process_1.spawn(command.name, command.args, Object.assign(Object.assign({}, options), { env }));
        // TODO use proper signal typing, see child exit cb types
        const result = yield new Promise((resolve, reject) => {
            // NOTE "exit" may fire after "error", in which case it will be a noop
            // as per how promises work.
            // When spawn is executed in pipe mode, then we buffer up the data for
            // later inspection
            // TODO return type should use conditional types to express mapping
            // between stdio option settings and resulting returned std err/out buffers.
            let stderr = null;
            let stdout = null;
            if (child.stderr) {
                stderr = '';
                child.stderr.on('data', bufferStderr);
            }
            if (child.stdout) {
                stdout = '';
                child.stdout.on('data', bufferStdout);
            }
            function bufferStderr(chunk) {
                stderr += String(chunk);
            }
            function bufferStdout(chunk) {
                stdout += String(chunk);
            }
            child.once('error', error => {
                var _a;
                const richError = createCommandError({
                    command: commandRaw,
                    underlyingError: error,
                    stderr,
                    stdout,
                    signal: null,
                    exitCode: null,
                });
                if (((_a = options) === null || _a === void 0 ? void 0 : _a.require) === true) {
                    cleanup();
                    reject(richError);
                }
                else {
                    cleanup();
                    resolve({
                        command: commandRaw,
                        stdout,
                        stderr,
                        signal: null,
                        error: richError,
                        exitCode: null,
                    });
                }
            });
            child.once('exit', (exitCode, signal) => {
                var _a;
                const error = isFailedExitCode(exitCode)
                    ? createCommandError({
                        command: commandRaw,
                        underlyingError: null,
                        signal,
                        stderr,
                        stdout,
                        exitCode,
                    })
                    : null;
                if (((_a = options) === null || _a === void 0 ? void 0 : _a.require) === true && isFailedExitCode(exitCode)) {
                    cleanup();
                    reject(error);
                }
                else {
                    cleanup();
                    resolve({
                        command: commandRaw,
                        signal,
                        stderr,
                        stdout,
                        exitCode,
                        error,
                    });
                }
            });
            function cleanup() {
                var _a, _b;
                (_a = child.stderr) === null || _a === void 0 ? void 0 : _a.removeListener('data', bufferStderr);
                (_b = child.stdout) === null || _b === void 0 ? void 0 : _b.removeListener('data', bufferStdout);
            }
        });
        return result;
    });
}
exports.run = run;
/**
 * Util that binds a command to run making it easy to abstract a parent command.
 *
 * @example
 *
 *    const git = createRunner('git')
 *    const result = await git('status')
 */
exports.createRunner = (cwd) => {
    return (cmd, opts) => {
        return run(cmd, Object.assign(Object.assign({}, opts), { cwd }));
    };
};
/**
 * Create an error enriched with properties related to the run result.
 */
function createCommandError({ command, signal, stderr, stdout, exitCode, underlyingError, }) {
    const error = new Error(common_tags_1.stripIndent `
    The following command failed to complete successfully:

        ${command}

    It ended with this exit code:

        ${exitCode}

    This underlying error occured (null = none occured):

        ${underlyingError}

    It received signal (null = no signal received):

        ${signal}

    It output on stderr (null = not spawned in pipe mode):

        ${stderr}

    It output on stdout (null = not spawned in pipe mode):

        ${stdout}
  `);
    // @ts-ignore
    error.exitCode = exitCode;
    // @ts-ignore
    error.signal = signal;
    // @ts-ignore
    error.stderr = stderr;
    // @ts-ignore
    error.stdout = stdout;
    return error;
}
/**
 * Convert a command string into something the standard lib child process module
 * APIs will accept.
 */
function parseCommandString(cmd) {
    const [name, ...args] = cmd.split(' ');
    return {
        name,
        args,
    };
}
/**
 * Is the given exit code a failure or success? Intended for internal use,
 * handles null which is convenient for this module.
 */
function isFailedExitCode(exitCode) {
    return typeof exitCode === 'number' && exitCode !== 0;
}
/**
 * Check if this process was created from the bin of the given project or not.
 * @param packageJsonPath
 */
function isProcessFromProjectBin(packageJsonPath) {
    const processBinPath = process.argv[1];
    const processBinDirPath = Path.dirname(processBinPath);
    const projectBinDirPath = Path.join(Path.dirname(packageJsonPath), 'node_modules/.bin');
    return processBinDirPath !== projectBinDirPath;
}
exports.isProcessFromProjectBin = isProcessFromProjectBin;
/**
 * Log a meaningful semantic error message sans stack track and then crash
 * the program with exit code 1. Parameters are a passthrough to `console.error`.
 */
function fatal(format, ...vars) {
    console.error(format, ...vars);
    process.exit(1);
}
exports.fatal = fatal;
