"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Calculate the stable increment to a given version.
 */
function incStable(bumpType, v) {
    const { vprefix, major, minor, patch } = v;
    switch (bumpType) {
        case 'major':
            return createStable(major + 1, 0, 0, { vprefix });
        case 'minor':
            return createStable(major, minor + 1, 0, { vprefix });
        case 'patch':
            return createStable(major, minor, patch + 1, { vprefix });
    }
}
exports.incStable = incStable;
/**
 * Add pre-release info to a stable release. In other words convert a stable
 * release into a pre-release one.
 */
function stableToPreview(v, identifier, buildNum) {
    return createPreRelease(v.major, v.minor, v.patch, identifier, buildNum);
}
exports.stableToPreview = stableToPreview;
/**
 * Create a semantic pre-release version programatically.
 */
function createPreRelease(major, minor, patch, identifier, buildNum, opts) {
    var _a, _b;
    return {
        version: `${major}.${minor}.${patch}-${identifier}.${buildNum}`,
        vprefix: (_b = (_a = opts) === null || _a === void 0 ? void 0 : _a.vprefix, (_b !== null && _b !== void 0 ? _b : false)),
        major,
        minor,
        patch,
        preRelease: {
            buildNum,
            identifier,
        },
    };
}
exports.createPreRelease = createPreRelease;
/**
 * Create a semantic version programatically.
 */
function createStable(major, minor, patch, opts) {
    var _a, _b;
    return {
        major,
        minor,
        patch,
        vprefix: (_b = (_a = opts) === null || _a === void 0 ? void 0 : _a.vprefix, (_b !== null && _b !== void 0 ? _b : false)),
        version: `${major}.${minor}.${patch}`,
    };
}
exports.createStable = createStable;
/**
 * Is the given version a preview one?
 */
function isPreview(v) {
    return v.preRelease !== undefined;
}
exports.isPreview = isPreview;
/**
 * Is the given version a stable one?
 */
function isStable(v) {
    return !isPreview(v);
}
exports.isStable = isStable;
/**
 * Parse a version that you believe should be a preview variant. If not, an error is thrown.
 */
function parsePreview(ver) {
    const result = parse(ver);
    if (result === null)
        return null;
    if (result.preRelease) {
        return result;
    }
    throw new Error(`Given version string ${ver} could not be parsed as a preview.`);
}
exports.parsePreview = parsePreview;
/**
 * Parse a version string into structured data.
 */
function parse(ver) {
    const result = ver.match(/^(v)?(\d+).(\d+).(\d+)$|^(v)?(\d+).(\d+).(\d+)-(\w+).(\d+)$/);
    if (result === null)
        return null;
    if (result[6]) {
        const vprefix = result[5] === 'v';
        const major = parseInt(result[6], 10);
        const minor = parseInt(result[7], 10);
        const patch = parseInt(result[8], 10);
        const identifier = result[9];
        const buildNum = parseInt(result[10], 10);
        return {
            version: `${major}.${minor}.${patch}-${identifier}.${buildNum}`,
            vprefix,
            major,
            minor,
            patch,
            preRelease: {
                identifier,
                buildNum,
            },
        };
    }
    const vprefix = result[1] === 'v';
    const major = parseInt(result[2], 10);
    const minor = parseInt(result[3], 10);
    const patch = parseInt(result[4], 10);
    return {
        version: `${major}.${minor}.${patch}`,
        vprefix,
        major,
        minor,
        patch,
    };
}
exports.parse = parse;
function setBuildNum(ver, buildNum) {
    return Object.assign(Object.assign({}, ver), { preRelease: Object.assign(Object.assign({}, ver.preRelease), { buildNum }), version: `${ver.major}.${ver.minor}.${ver.patch}-${ver.preRelease.identifier}.${buildNum}` });
}
exports.setBuildNum = setBuildNum;
exports.zeroVer = createStable(0, 0, 0);
exports.zeroBuildNum = 0;
