"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const errors_1 = require("@oclif/errors");
/**
 *
 * @example
 *
 * ```ts
 * guard(ctx)
 *   .must(isTrunk())
 *   .check(flags.upToDate ? 'must' : 'prefer', upToDate())
 *
 * if (flags.npmPublish) {
 *   guard.must(publishRights())
 * }
 *
 * guard.run()
 * ```
 */
function check(options) {
    const contextualizedCheckers = [];
    const api = {
        check(level, validator) {
            contextualizedCheckers.push({ level, validator });
            return api;
        },
        stopUnless(checker) {
            return api.check('prefer', checker);
        },
        errorUnless(checker) {
            return api.check('must', checker);
        },
        run() {
            return contextualizedCheckers.reduce((results, requirement) => {
                const resultSugar = requirement.validator.run(options.context);
                const result = typeof resultSugar === 'boolean'
                    ? booleanResult(resultSugar)
                    : resultSugar;
                const group = result.kind === 'pass'
                    ? results.passes
                    : requirement.level === 'must'
                        ? results.errors
                        : results.stops;
                group.push({
                    details: result.details,
                    summary: requirement.validator.summary,
                    code: requirement.validator.code,
                });
                return results;
            }, createReport());
        },
    };
    return api;
}
exports.check = check;
/**
 *
 */
function createReport() {
    return {
        stops: [],
        errors: [],
        passes: [],
    };
}
/**
 * Enforce the result of a report. Throws errors/stop-signal that will halt the
 * CLI if needed.
 */
function guard(input) {
    if (input.report.errors.length) {
        if (input.json) {
            if (input.report.errors.length) {
                throw new JSONCLIError({
                    context: input.context,
                    failures: input.report.errors,
                });
            }
        }
        else {
            throw new PassthroughCLIError(input.report.errors
                .map(failure => {
                return failure.summary;
            })
                .join('\n - '));
        }
    }
    if (input.report.stops.length) {
        throw new CLIStop({ reasons: input.report.stops, json: input.json });
    }
}
exports.guard = guard;
//
// Errors for halting CLI process
//
class JSONCLIError extends errors_1.CLIError {
    constructor(errorObject) {
        super('...todo...', { exit: 1 });
        this.errorObject = errorObject;
        this.code = 'JSONError';
    }
    render() {
        return JSON.stringify(this.errorObject);
    }
}
exports.JSONCLIError = JSONCLIError;
class PassthroughCLIError extends errors_1.CLIError {
    constructor(errorString) {
        super('...todo...', { exit: 1 });
        this.errorString = errorString;
        this.code = 'PassthroughError';
    }
    render() {
        return this.errorString;
    }
}
exports.PassthroughCLIError = PassthroughCLIError;
class CLIStop extends errors_1.CLIError {
    constructor(input) {
        super('...todo...', { exit: 0 });
        this.input = input;
        this.code = 'Stop';
    }
    render() {
        if (this.input.json) {
            return JSON.stringify({ reasons: this.input.reasons });
        }
        else {
            return ('Nothing to do:\n\n' +
                this.input.reasons.map(r => r.summary).join('\n -'));
        }
    }
}
exports.CLIStop = CLIStop;
function booleanResult(bool) {
    return bool ? { kind: 'pass', details: {} } : { kind: 'fail', details: {} };
}
exports.booleanResult = booleanResult;
