"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const util_1 = require("util");
const utils_1 = require("../lib/utils");
/**
 * Passthrough to the underlying output mechanism.
 */
function outputRaw(message) {
    console.log(message);
}
exports.outputRaw = outputRaw;
/**
 * This module encapsulates the various structures that dripip may output.
 * Precise data details stay local throughout the codebase but core concerns
 * like data envelop and serialization are encoded here.
 */
/**
 * Output Ok data to stdout. Use this to handle general output.
 */
function outputOk(type, data) {
    outputMessage(createOk(type, data));
}
exports.outputOk = outputOk;
/**
 * Outout Exception data to stdout. Unlike an error, exceptions are failure
 * scenarios that are known to be possible, and thus handled gracefully.
 *
 * @param identifier This is a meaningful exception slug like `no_permissions`.
 * Use this for to organize exeptions into a catalogue.
 *
 * @param summary Free form text that briefly explains what the exception
 * is, why it is happening, etc. Do not rely on rich context rendering here.
 * Instead prefer to put the data/variables at hand into the context object.
 *
 * @param context Arbitrary data that relates to the exception at hand.
 */
function outputException(identifier, summary, opts) {
    output(createException(identifier, { summary, context: opts.context }), {
        json: opts.json,
    });
}
exports.outputException = outputException;
function output(message, opts) {
    if (opts.json) {
        outputMessage(message);
    }
    else {
        if (message.kind === 'exception') {
            let s = '';
            s += `An exception occured: ${message.type}\n`;
            s += `\n`;
            s += message.data.summary;
            if (message.data.context &&
                Object.keys(message.data.context).length > 0) {
                s += `\n`;
                s += util_1.format('%j', message.data.context);
            }
            outputRaw(s);
        }
        else if (message.kind === 'ok') {
            let s = '';
            s += message.data;
        }
        else {
            utils_1.casesHandled(message);
        }
    }
}
exports.output = output;
/**
 * See output version docs.
 */
function createOk(type, data) {
    return {
        kind: 'ok',
        type,
        data,
    };
}
exports.createOk = createOk;
/**
 * See output version docs.
 */
function createException(type, { summary, context, }) {
    return {
        kind: 'exception',
        type,
        data: {
            summary,
            context: (context !== null && context !== void 0 ? context : {}),
        },
    };
}
exports.createException = createException;
function didNotPublish(input) {
    outputJson({
        kind: 'did_not_publish',
        reasons: input.reasons,
    });
}
exports.didNotPublish = didNotPublish;
function didPublish(input) {
    outputJson(Object.assign({ kind: 'did_publish' }, input.release));
}
exports.didPublish = didPublish;
/**
 * Output Message to stdout.
 */
function outputMessage(msg) {
    outputJson(msg);
}
exports.outputMessage = outputMessage;
/**
 * Output JSON to stdout.
 */
function outputJson(msg) {
    process.stdout.write(JSON.stringify(msg));
}
exports.outputJson = outputJson;
